==> m4e_sdk-0.0.3-dev.13.tar.gz: lib/src/data/api/auth_api.dart <==
import 'dart:convert';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter_secure_storage/flutter_secure_storage.dart' as storage;
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../m4e_forms/m4e_forms.dart';
import '../../models/models.dart';

/// Accessing the instantiated `AuthApi` methods
/// is done by passing its type to the GetIt instance
///
/// Example use of `AuthApi`:
/// ```
/// import 'package:m4edart/m4e_sdk_initialize.dart' as sdk;
///
/// sdk.m4e<AuthApi>().getCurrentUser();
/// ```
///
abstract class AuthApi {
  /// Returns cached user authentication credentials `M4eAuthPhoneNumberForm`
  ///
  /// Throws `AuthException` with code `NO_CACHED_AUTH_CREDENTIAL`
  /// when no user is authenticated
  Future<M4eAuthPhoneNumberForm> getUserCredentialsFromCache();

  /// Cache user m4e authentication credential after a successful authentication
  ///
  Future<void> cacheUserCredentials(M4eAuthPhoneNumberForm credential);

  /// Retrives cached authentication token
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  Future<M4eAuthToken> getAuthTokenFromCache();

  /// Cache m4e token returned after authentication
  ///
  Future<void> cacheAuthToken(M4eAuthToken authToken);

  /// Returns m4e access-token with `refresh-token`
  ///
  /// Throws `ServerException` when server call fails or refresh token is invalid
  Future<M4eAuthToken> refreshAccessToken(M4eAuthToken authToken);

  /// Sends OTP code to device with phone number paramter
  ///
  /// Returns `user-id` of phone-number sent an OTP
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  Future<M4eUniqueId> verifyPhoneNumber(
      M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm);

  /// Verifies OTP code sent to device.
  ///
  /// Returns `user-id` of phone-number sent an OTP
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  Future<M4eAuthToken> verifyOTPCode(
    M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm,
    M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm,
  );

  /// Signout authenticated m4e user
  Future<void> signout();
}

class AuthApiImpl implements AuthApi {
  final Dio _client;
  final storage.FlutterSecureStorage _storage;

  AuthApiImpl({
    @required Dio client,
    @required storage.FlutterSecureStorage storage,
    @required M4eLoggingInterceptors loggingInterceptor,
  })  : assert(client != null),
        assert(storage != null),
        assert(loggingInterceptor != null),
        _client = client,
        _storage = storage {
    // logs requests
    _client?.interceptors?.add(loggingInterceptor);
  }

  String base64EncodedChallenge(
    M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm,
    M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm,
  ) {
    final _headerParams =
        '${m4eAuthPhoneNumberForm.id.toString()}:${m4eAuthOTPVerificationForm.otpCode}:${DateTime.now().toTimestampSinceEpoch()}';

    // prints header in debug mode
    logger('Base64 encoded header $_headerParams');
    return base64.encode(_headerParams.codeUnits);
  }

  Future<M4eAuthPhoneNumberForm> getUserCredentialsFromCache() async {
    final _m4eAuthPhoneNumberCredentialString =
        await _storage.read(key: M4ePersistence.kM4eAuthCredential);

    if (_m4eAuthPhoneNumberCredentialString != null) {
      return M4eAuthPhoneNumberForm.fromJson(
          json.decode(_m4eAuthPhoneNumberCredentialString));
    }

    throw M4eExceptionMessages.kNoAuthCredential;
  }

  Future<void> cacheUserCredentials(M4eAuthPhoneNumberForm credential) async {
    final _m4eAuthPhoneNumberCredentialString =
        json.encode(credential.toJson());

    return await _storage.write(
        key: M4ePersistence.kM4eAuthCredential,
        value: _m4eAuthPhoneNumberCredentialString);
  }

  Future<M4eAuthToken> getAuthTokenFromCache() async {
    final _authTokenString =
        await _storage.read(key: M4ePersistence.kM4eAuthToken);

    if (_authTokenString != null) {
      return M4eAuthToken.fromJson(json.decode(_authTokenString));
    }

    throw M4eExceptionMessages.kNoAuthUser;
  }

  Future<void> cacheAuthToken(M4eAuthToken authToken) async {
    final _authTokenString = json.encode(authToken.toJson());

    return await _storage.write(
        key: M4ePersistence.kM4eAuthToken, value: _authTokenString);
  }

  // REMOTE===========================================================================================

  Future<M4eAuthToken> refreshAccessToken(M4eAuthToken authToken) async {
    // TODO: implement refresh-token scheme
    return authToken;
  }

  Future<M4eUniqueId> verifyPhoneNumber(
      M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm) async {
    assert(m4eAuthPhoneNumberForm != null,
        'M4eAuthPhoneNumberForm should not be null');
    try {
      final _response = await _client.put('/auth/claim',
          data: json.encode(m4eAuthPhoneNumberForm.toJson()));

      if (_response.statusCode == 200) {
        return M4eUniqueId.fromJson(_response.data);
      }

      throw M4eExceptionMessages.kFailedTransactionException;
    } on DioError catch (e) {
      // print error in console in debug mode
      logger(e);

      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
        }
        if (e.response.statusCode == 404) {
          throw M4eExceptionMessages.kNotFoundException;
        }
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw M4eExceptionMessages.kUndefinedException;
    }
  }

  Future<M4eAuthToken> verifyOTPCode(
    M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm,
    M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm,
  ) async {
    assert(m4eAuthOTPVerificationForm != null,
        'M4eAuthOTPVerificationForm should not be null');
    try {
      final _response = await _client.put(
        '/auth/verify',
        data: json.encode(m4eAuthPhoneNumberForm.toJson()),
        options: Options(
          headers: {
            HttpHeaders.authorizationHeader:
                'Challenge ${base64EncodedChallenge(m4eAuthPhoneNumberForm, m4eAuthOTPVerificationForm)}'
          },
        ),
      );

      if (_response.statusCode == 200) {
        return M4eAuthToken.fromJson(_response.data);
      }

      throw M4eExceptionMessages.kFailedTransactionException;
    } on DioError catch (e) {
      // print error in console in debug mode
      logger(e);

      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
        }
        if (e.response.statusCode == 404) {
          throw M4eExceptionMessages.kNotFoundException;
        }
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw M4eExceptionMessages.kUndefinedException;
    }
  }

  Future<void> signout() async {
    return await _storage.deleteAll();
  }
}

==> m4e_sdk-0.0.4-dev.10.tar.gz: lib/src/data/api/auth_api.dart <==
import 'dart:convert';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter_secure_storage/flutter_secure_storage.dart' as storage;
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../m4e_forms/m4e_forms.dart';
import '../../models/models.dart';

/// Accessing the instantiated `AuthApi` methods
/// is done by passing its type to the GetIt instance
///
/// Example use of `AuthApi`:
/// ```
/// import 'package:m4edart/m4e_sdk_initialize.dart' as sdk;
///
/// sdk.m4e<AuthApi>().getCurrentUser();
/// ```
///
abstract class AuthApi {
  /// Returns cached user authentication credentials `M4eAuthPhoneNumberForm`
  ///
  /// Throws `AuthException` with code `NO_CACHED_AUTH_CREDENTIAL`
  /// when no user is authenticated
  Future<M4eAuthPhoneNumberForm> getUserCredentialsFromCache();

  /// Cache user m4e authentication credential after a successful authentication
  ///
  Future<void> cacheUserCredentials(M4eAuthPhoneNumberForm credential);

  /// Retrives cached authentication token
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  Future<M4eAuthToken> getAuthTokenFromCache();

  /// Cache m4e token returned after authentication
  ///
  Future<void> cacheAuthToken(M4eAuthToken authToken);

  /// Returns m4e access-token with `refresh-token`
  ///
  /// Throws `ServerException` when server call fails or refresh token is invalid
  Future<M4eAuthToken> refreshAccessToken(M4eAuthToken authToken);

  /// Sends OTP code to device with phone number paramter
  ///
  /// Returns `user-id` of phone-number sent an OTP
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  Future<M4eUniqueId> verifyPhoneNumber(
      M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm);

  /// Verifies OTP code sent to device.
  ///
  /// Returns `user-id` of phone-number sent an OTP
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  Future<M4eAuthToken> verifyOTPCode(
    M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm,
    M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm,
  );

  /// Signout authenticated m4e user
  Future<void> signout();
}

class AuthApiImpl implements AuthApi {
  final Dio _client;
  final storage.FlutterSecureStorage _storage;

  AuthApiImpl({
    @required Dio client,
    @required storage.FlutterSecureStorage storage,
    @required M4eLoggingInterceptors loggingInterceptor,
  })  : assert(client != null),
        assert(storage != null),
        assert(loggingInterceptor != null),
        _client = client,
        _storage = storage {
    // logs requests
    _client?.interceptors?.add(loggingInterceptor);
  }

  String base64EncodedChallenge(
    M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm,
    M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm,
  ) {
    final _headerParams =
        '${m4eAuthPhoneNumberForm.id.toString()}:${m4eAuthOTPVerificationForm.otpCode}:${DateTime.now().toTimestampSinceEpoch()}';

    // prints header in debug mode
    return base64.encode(_headerParams.codeUnits);
  }

  Future<M4eAuthPhoneNumberForm> getUserCredentialsFromCache() async {
    final _m4eAuthPhoneNumberCredentialString =
        await _storage.read(key: M4ePersistence.kM4eAuthCredential);

    if (_m4eAuthPhoneNumberCredentialString != null) {
      return M4eAuthPhoneNumberForm.fromJson(
          json.decode(_m4eAuthPhoneNumberCredentialString));
    }

    throw M4eExceptionMessages.kNoAuthCredential;
  }

  Future<void> cacheUserCredentials(M4eAuthPhoneNumberForm credential) async {
    final _m4eAuthPhoneNumberCredentialString =
        json.encode(credential.toJson());

    return await _storage.write(
        key: M4ePersistence.kM4eAuthCredential,
        value: _m4eAuthPhoneNumberCredentialString);
  }

  Future<M4eAuthToken> getAuthTokenFromCache() async {
    final _authTokenString =
        await _storage.read(key: M4ePersistence.kM4eAuthToken);

    if (_authTokenString != null) {
      return M4eAuthToken.fromJson(json.decode(_authTokenString));
    }

    throw M4eExceptionMessages.kUnauthorized;
  }

  Future<void> cacheAuthToken(M4eAuthToken authToken) async {
    final _authTokenString = json.encode(authToken.toJson());

    return await _storage.write(
        key: M4ePersistence.kM4eAuthToken, value: _authTokenString);
  }

  // REMOTE===========================================================================================

  Future<M4eAuthToken> refreshAccessToken(M4eAuthToken authToken) async {
    // TODO: implement refresh-token scheme
    return authToken;
  }

  Future<M4eUniqueId> verifyPhoneNumber(
      M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm) async {
    assert(m4eAuthPhoneNumberForm != null,
        'M4eAuthPhoneNumberForm should not be null');
    try {
      final _response = await _client.put('/auth/claim',
          data: json.encode(m4eAuthPhoneNumberForm.toJson()));

      return M4eUniqueId.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }

  Future<M4eAuthToken> verifyOTPCode(
    M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm,
    M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm,
  ) async {
    assert(m4eAuthOTPVerificationForm != null,
        'M4eAuthOTPVerificationForm should not be null');
    try {
      final _response = await _client.put(
        '/auth/verify',
        data: json.encode(m4eAuthPhoneNumberForm.toJson()),
        options: Options(
          headers: {
            HttpHeaders.authorizationHeader:
                'Challenge ${base64EncodedChallenge(m4eAuthPhoneNumberForm, m4eAuthOTPVerificationForm)}'
          },
        ),
      );

      return M4eAuthToken.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }

  Future<void> signout() async {
    return await _storage.deleteAll();
  }
}
